/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* 
 * Handle the table of declared peers.
 *
 * Since we always search the score for a given peer, the internal data is a hash table of the peers, and for 
 * each entry the list of realms it is configured for.
 */

#include "rt_default.h"

#include <assert.h>

#ifndef RTD_HASH_SIZE
#define RTD_HASH_SIZE 4
#endif /* RTD_HASH_SIZE */

/* This is the list of realms served by a given peer. */
typedef struct _rtd_rlm {
	char 		*realm; /* the realm name */
	size_t		 len;	/* realm string length */
	int 		 pref;  /* is the peer a prefered route ? 0: no, 1 : yes */
	struct _rtd_rlm *next;  /* next realm, or NULL for the last one. The list is ordered alphabetically */
} rtd_rlm_t;

/* This is the definition of one peer */
typedef struct _rtd_peer {
	uint32_t	 hash; 		/* the hash value of the peer */
	char		*peer; 		/* peer diameter id */
	int		 defroute; 	/* -1: not a default route, 0: yes, 1: yes, prefered */
	rtd_rlm_t	 realms; 	/* sentinel for the list of realms served by this peer */
	struct _rtd_peer *next;		/* next peer in the sublist. List is ordered by hash and peer id */
} rtd_peer_t;

static rtd_peer_t RTD_HASH[ 1 << RTD_HASH_SIZE ]; /* The hash table. No lock is needed since we are not multithreaded when doing write accesses */
#define H_MASK( __hash) ((__hash) & (( 1 << RTD_HASH_SIZE ) - 1))
#define H_LINE( __hash) (&RTD_HASH[H_MASK( __hash)])

/* We use the same hash function as the main daemon */
/********************************************************************************************************/
/* Hash function -- credits to Austin Appleby, thank you ^^ */
/* See http://murmurhash.googlepages.com for more information on this function */

/* the strings are always aligned properly, so we use the simple MurmurHash2 function */
#define _HASH_MIX(h,k,m) { k *= m; k ^= k >> r; k *= m; h *= m; h ^= k; }
static uint32_t rtd_hash ( char * string)
{
	size_t len = strlen(string);
	uint32_t hash = len;
	char * data = string;
	
	const unsigned int m = 0x5bd1e995;
	const int r = 24;
	
	/* Check the alignment is really correct -- otherwise we have to switch to the other version */
	assert(((long)string & 3) == 0);

	while(len >= 4)
	{
		/* Mix 4 bytes at a time into the hash */
		uint32_t k = *(uint32_t *)data;	/* We don't care about the byte order */
		
		_HASH_MIX(hash, k, m);

		data += 4;
		len -= 4;
	}
	
	/* Handle the last few bytes of the input */
	switch(len) {
		case 3: hash ^= data[2] << 16;
		case 2: hash ^= data[1] << 8;
		case 1: hash ^= data[0];
	        	hash *= m;
	}

	/* Do a few final mixes of the hash to ensure the last few
	   bytes are well-incorporated. */
	hash ^= hash >> 13;
	hash *= m;
	hash ^= hash >> 15;

	return hash;
} 
/********************************************************************************************************/

/* Initialize the table */
int rtd_init(void)
{
	memset(&RTD_HASH, 0, sizeof(RTD_HASH));
	return 0;
}

/* End the table */
void rtd_fini(void)
{
	int i;
	
	/* Empty the hash table */
	for (i = 0; i < sizeof(RTD_HASH) / sizeof(RTD_HASH[0]); i++) {
		while (RTD_HASH[i].next != NULL) {
			rtd_peer_t * peer = RTD_HASH[i].next;
			RTD_HASH[i].next = peer->next;
			
			while (peer->realms.next != NULL) {
				rtd_rlm_t * realm = peer->realms.next;
				peer->realms.next = realm->next;
				
				free(realm->realm);
				free(realm);
			}
			
			free(peer->peer);
			free(peer);
		}
	}
}

/* Add a new entry */
int rtd_add(char * realm, char * peer, int prefered)
{
	uint32_t hash;
	rtd_peer_t * peerdef = NULL;
	rtd_rlm_t * realmdef = NULL;
	int match = 0;
	
	TRACE_ENTRY("%p %p %d", realm, peer, prefered);
	
	if (peer == NULL) {
		TRACE_DEBUG(INFO, "Invalid parameter");
		return EINVAL;
	}
	
	hash = rtd_hash(peer);
	
	/* Now search if we already have this peer, or its previous element otherwise */
	for (peerdef = H_LINE(hash); peerdef->next != NULL; peerdef = peerdef->next) {
		int cmp;
		if (peerdef->next->hash < hash)
			continue;
		if (peerdef->next->hash > hash)
			break;
		cmp = strcasecmp(peerdef->next->peer, peer);
		if (cmp < 0)
			continue;
		if (cmp == 0)
			match = 1;
		break;
	}
	
	if (match) {
		free(peer);
	} else {
		rtd_peer_t * new = (rtd_peer_t *)malloc(sizeof(rtd_peer_t));
		if (new == NULL) {
			log_error("Unable to store a route information: %s\n", strerror(errno));
			return ENOMEM;
		}
		memset(new, 0, sizeof(rtd_peer_t));
		new->hash = hash;
		new->peer = peer;
		new->defroute = -1;
		new->next = peerdef->next;
		peerdef->next = new;
	}
	
	peerdef = peerdef->next;
	
	/* Now handle the realm part */
	
	if (realm == NULL) {
		if (prefered) {
			peerdef->defroute = 1;
			/* might check ancient value was -1 otherwise multiple definitions */
		} else {
			if (peerdef->defroute == -1)
				peerdef->defroute = 0;
			/* else: multiple definitions of the same peer... not a problem */
		}
		
		return 0;
	}
	
	match = 0;
	
	/* search for this realm in the list, or stop at previous realm to create a new entry */
	for (realmdef = &peerdef->realms; realmdef->next != NULL; realmdef = realmdef->next) {
		int cmp = strcasecmp(realmdef->next->realm, realm);
		if (cmp < 0)
			continue;
		if (cmp == 0)
			match = 1;
		break;
	}
	
	if (match) {
		free(realm);
	} else {
		rtd_rlm_t * new = (rtd_rlm_t *)malloc(sizeof(rtd_rlm_t));
		if (new == NULL) {
			log_error("Unable to store a route information: %s\n", strerror(errno));
			return ENOMEM;
		}
		memset(new, 0, sizeof(rtd_rlm_t));
		new->realm = realm;
		new->len = strlen(realm);
		new->next = realmdef->next;
		realmdef->next = new;
	}
	
	if (prefered)
		realmdef->next->pref = 1;
		
	return 0;
}

/* Dump all entries */
void rtd_dump(void)
{
	int i;
	
	TRACE_DEBUG(FULL, " Routing Table (by destination peer):");
	for (i = 0; i < sizeof(RTD_HASH) / sizeof(RTD_HASH[0]); i++) {
		rtd_peer_t * peer;
		for (peer = RTD_HASH[i].next; peer != NULL; peer=peer->next) {
			rtd_rlm_t * realm;
			
			TRACE_DEBUG(FULL, "  ... '%s':", peer->peer);
			if (peer->defroute != -1) {
				TRACE_DEBUG(FULL, "      ... default route%s", peer->defroute?" *":"");
			}
			
			for (realm = peer->realms.next; realm != NULL; realm = realm->next) {
				TRACE_DEBUG(FULL, "      ... '%s'%s", realm->realm, realm->pref?" *":"");
			}
		}
	}
}

/* Search an entry */
int rtd_search( unsigned char * realm, size_t realm_len, char * peer, int * score)
{
	uint32_t hash;
	rtd_peer_t * pdef;
	rtd_rlm_t * rdef;
	int match = 0;
	
	TRACE_ENTRY("%p %d %p %p", realm, realm_len, peer, score);
	
	assert( peer && score );
	
	/* Default: no information */
	*score = rtd_conf->sc_no_info;
	
	/* Search if we have this peer definition */
	hash = rtd_hash(peer);
	
	/* Now search if we have this peer */
	for (pdef = H_LINE(hash)->next; pdef != NULL; pdef = pdef->next) {
		int cmp;
		if (pdef->hash < hash)
			continue;
		if (pdef->hash > hash)
			break;
		cmp = strcasecmp(pdef->peer, peer);
		if (cmp < 0)
			continue;
		if (cmp == 0)
			match = 1;
		break;
	}
	
	if (!match)
		goto end;
	
	/* Is the peer defined as a default route? */
	switch (pdef->defroute) {
		case -1:
			/* no information, we already saved this value */
			break;
		case 0:
			*score = rtd_conf->sc_all;
			break;
		case 1:
			*score = rtd_conf->sc_all_pref;
			break;
	}
	
	/* Now handle the realm part */
	if (realm == NULL)
		goto end;
	
	match = 0;
	
	/* search for this realm in the list */
	for (rdef = pdef->realms.next; rdef != NULL; rdef = rdef->next) {
		int cmp = strncasecmp(rdef->realm, (char *)realm, realm_len);
		if (cmp < 0)
			continue;
		if (cmp > 0)
			break;
		if (rdef->len < realm_len)
			continue;
		if (rdef->len == realm_len)
			match = 1;
		break;
	}
	
	if (!match)
		goto end;
	
	if (rdef->pref)
		*score = rtd_conf->sc_rlm_pref;
	else
		*score = rtd_conf->sc_rlm;
	
end:
	TRACE_DEBUG(FULL, "table: '%s' -> %d", peer, *score);
	return 0;
}

