/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the functions to manage the _peer_t structures and the peer hash table.
 *
 */
#include "waaad-internal.h"
#include "peer-internal.h"


/* The hash table */
_peer_hash_t 	_peer_hash[1 << _PEER_HASH_SIZE];

/* The list of active peers */
uti_list_t 	_peers_actives;


/* Initialize the hash table */
int _peer_struct_init()
{
	int i;
	
	uti_list_init(&_peers_actives, NULL);
	
	/* Initialize the hash table */
	memset(_peer_hash, 0, sizeof(_peer_hash));
	
	for (i = 0; i < sizeof(_peer_hash) / sizeof(_peer_hash[0]); i++) {
		uti_list_init(&_peer_hash[i].all, 	&_peer_hash[i]);
		CHECK_POSIX(  pthread_mutex_init(&_peer_hash[i].lock, NULL)  );
	}
	
	return 0;
}

/* Destroy the hash table and all peers */
int _peer_struct_fini()
{
	int i;
	
	for (i = 0; i < sizeof(_peer_hash) / sizeof(_peer_hash[0]); i++) {
		while ( ! IS_LIST_EMPTY( &_peer_hash[i].all )  ) {
			_peer_t * peer = _P(_peer_hash[i].all.next->o);
			CHECK_FCT_DO(_peer_psm_stop(peer, -1), /* continue */ );
			CHECK_FCT_DO(_peer_struct_unlink(peer), /* continue */ );
			CHECK_FCT_DO(_peer_struct_destroy(&peer), /* continue */ );
		}
		CHECK_POSIX_DO(  pthread_mutex_destroy(&_peer_hash[i].lock), /* continue */  );
	}
	
	/* done */
	return 0;
}

/* Add an element in a peer_appl_t list of a peer */
int _peer_struct_appl_add( _peer_t * peer, application_id_t a, vendor_id_t v, peer_app_type_t t)
{
	int i;
	
	if (a == 0xffffffff) {
		peer->p_app_relay = 1;
		return 0;
	}
	
	CHECK_MALLOC( peer->p_app_list = (peer_appl_t *)realloc(peer->p_app_list, (peer->p_app_size + 2) * sizeof(peer_appl_t)) );
	peer->p_app_size += 1;

	/* We must move the elements after the one we are inserting */
	/* ( a b c d e 0 ) --> ( a b N c d e 0 ) */

	memset( &peer->p_app_list[peer->p_app_size], 0, sizeof(peer_appl_t));

	i = peer->p_app_size - 2;

	while ((i >= 0) && (peer->p_app_list[i].a > a)) {
		memcpy( &peer->p_app_list[i + 1], &peer->p_app_list[i], sizeof(peer_appl_t));
		i--;
	}
	peer->p_app_list[i+1].a = a;
	peer->p_app_list[i+1].v = v;
	peer->p_app_list[i+1].t = t;
	
	return 0;
}
void _peer_struct_appl_reset( _peer_t * peer )
{
	free(peer->p_app_list);
	peer->p_app_list = NULL;
	peer->p_app_size = 0;
	peer->p_app_relay = 0;
	return;
}

/* Add an element in the p_remotesec list of a peer */
int _peer_struct_secid_add( _peer_t * peer, uint32_t secid)
{
	CHECK_MALLOC( peer->p_remotesec = (uint32_t *)realloc(peer->p_remotesec, (peer->p_remotesec_sz + 1) * sizeof(uint32_t)) );
	peer->p_remotesec[peer->p_remotesec_sz] = secid;
	peer->p_remotesec_sz += 1;
	return 0;
}
void _peer_struct_secid_reset( _peer_t * peer )
{
	free(peer->p_remotesec);
	peer->p_remotesec = NULL;
	peer->p_remotesec_sz = 0;
	return;
}
int _peer_struct_secid_check( _peer_t * peer, uint32_t secid)
{
	int i;
	
	for (i=0; i<peer->p_remotesec_sz; i++)
		if (peer->p_remotesec[i] == secid)
			return 1;
	
	return 0;
}

/* reset the remote addresses published in CER/CEA */
void _peer_struct_addrs_reset( _peer_t * peer )
{
	peer->p_peeraddr_sz = 0;
	free(peer->p_peeraddr);
	peer->p_peeraddr = NULL;
}

/* reset security module to default nosec */
void _peer_struct_secmod_reset(_peer_t * peer)
{
	if (peer->p_sec_hdl) {
		CHECK_FCT_DO(  sec_modunlink(peer->p_sec_hdl),  /* nothing */  );
		peer->p_sec_hdl = NULL;
	}
	
	peer->p_secmod = &_peer_secini_mod;
	
	while (!IS_LIST_EMPTY( &peer->p_sec_list ) )
	{
		uti_list_t * li = peer->p_sec_list.next;
		uti_list_unlink(li);
		CHECK_FCT_DO( sec_modunlink(((_sec_item_t *)(li))->hdl), /* continue */ );
		free(li);
	}
	
	return;
}

/* Create a new peer structure. It is not saved in the hash table */
/* id can be NULL or lower-case string */
int _peer_struct_malloc(char * id, _peer_t **new) 
{
	int ret = 0;
	char * diamid;
	_peer_t * peer;
	
	/* Ensure the location for the new peer is valid */
	CHECK_PARAMS( new );
	*new = NULL;
	
	/* The diameter id of the new peer */
	if (id) {
		diamid = id;
	} else {
		CHECK_MALLOC( diamid = strdup("(unknown peer)")  );
	}
	
	/* Now create the structure and initialize it */
	CHECK_MALLOC_DO( peer = _P(malloc(sizeof(_peer_t))), { ret = ENOMEM; goto rb1; } );
	
	/* Set everything to 0 */
	memset(peer, 0, sizeof(_peer_t));
	
	/* Initialize the fields */
	peer->p_eyec = PEER_EYEC;
	
	/* The hash will be computed later, only if the peer has to be inserted into the hash list */
	uti_list_init(&peer->p_global, peer);
	
	peer->p_diamid = diamid;
	/* The realm is NULL yet, it will be set later */
	peer->p_hbh = lrand48();
	
	/* p_flags are set to 0 */
	peer->p_state = STATE_DISABLED;
	uti_list_init(&peer->p_active, peer);
	/* p_psm will be created when the PSM is started */
	CHECK_POSIX_DO(  ret = pthread_mutex_init(&peer->p_lock, NULL),  goto rb2  );
	CHECK_POSIX_DO(  ret = pthread_cond_init(&peer->p_condvar, NULL),  goto rb3  );
	/* p_ts will be set later */
	uti_list_init(&peer->p_events, NULL);
	
	CHECK_FCT_DO(  meq_new(&peer->p_in_q),  goto rb4  );
	/* p_in_th will be created by the psm */
	CHECK_FCT_DO(  meq_new(&peer->p_out_q),  goto rb5  );
	/* p_out_th will be created by the psm */
	uti_list_init(&peer->p_sent, NULL);
	
	/* p_addinfo is set later */
	uti_list_init(&peer->p_exp_list, peer);
	
	/* All connection information will be set later */
	
	uti_list_init(&peer->p_sec_list, NULL);
	/* p_sec_hdl: no security handler initially */
	peer->p_secmod = &_peer_secini_mod;
	/* p_sec_session is initialized later when sending/receiving starts */
	
	/* The remaining of the structure is not set yet */
	
	/* Done: return 0 */
	*new = peer;
	return 0;
	
rb5:
	CHECK_FCT_DO(  meq_del(&peer->p_in_q), /* nothing */  );
rb4:
	CHECK_POSIX_DO(  pthread_cond_destroy(&peer->p_condvar), /* continue */  );
rb3:
	CHECK_POSIX_DO(  pthread_mutex_destroy(&peer->p_lock), /* continue */  );
rb2:
	free(peer);
rb1:
	free(diamid);
	return ret;
}

/* Remove a peer structure from all global lists. The peer must be in DISABLED state. */
int _peer_struct_unlink(_peer_t * peer)
{
	/* First, remove from expiry or deleted list if needed */
	if ( ! IS_LIST_EMPTY(&peer->p_exp_list) ) {
		CHECK_POSIX( pthread_mutex_lock(&_peers_lists_mtx) );
		uti_list_unlink(&peer->p_exp_list);
		CHECK_POSIX( pthread_mutex_unlock(&_peers_lists_mtx) );
	}
	
	/* Then remove from the hash table */
	if ( ! IS_LIST_EMPTY(&peer->p_global) ) {
		ASSERT( peer->p_hash ); /* Theoritically, the hash value may really be 0... Well, this is for debug, the ASSERT will not be compiled in final release. */
		CHECK_POSIX( pthread_mutex_lock(H_LOCK(peer->p_hash)) );
		uti_list_unlink(&peer->p_global);
		H_COUNT(peer->p_hash) -= 1;
		CHECK_POSIX( pthread_mutex_unlock(H_LOCK(peer->p_hash)) );
	}
	
	/* Done */
	return 0;
}

/* Destroy a peer structure. It must be already unlinked from any global list and in DISABLED state. Its lock must not be held. */
int _peer_struct_destroy(_peer_t ** peer)
{
	_peer_t * p;
	msg_t * msg = NULL;
	
	/* Check the parameter is acceptable */
	CHECK_PARAMS( peer && VALIDATE_PEER(*peer) );
	p=*peer;
	CHECK_PARAMS( (p->p_state == STATE_DISABLED) && IS_LIST_EMPTY(&p->p_global)  );
	
	/* Erase the location of the peer in the calling space */
	*peer = NULL;
	
	/* Some sanity checks */
	ASSERT( IS_LIST_EMPTY(&p->p_active)  );
	ASSERT( IS_LIST_EMPTY(&p->p_exp_list)  );
	ASSERT( p->p_psm == (pthread_t)NULL );
	ASSERT( p->p_ext_session == NULL );
	
	/* Now destroy the peer data */
	
	/* Free the strings */
	free(p->p_diamid);
	free(p->p_realm);
	
	/* Destroy the mutex and condvar */
	CHECK_POSIX_DO( pthread_mutex_destroy(&p->p_lock), /* continue */ );
	CHECK_POSIX_DO( pthread_cond_destroy(&p->p_condvar), /* continue */ );
	
	/* Destroy any pending event */
	_peer_events_reset(p);
	
	/* Handle any message that would remain in this peer */
	CHECK_FCT_DO( _peer_failover(p, 1), /* continue */  );
	
	/* for all messages in peer->p_in_q, destroy the message */
	while ( meq_tryget( p->p_in_q, &msg ) == 0 ) {
		TRACE_DEBUG(FULL, "Received message discarded during peer destruction (dump bellow)");
		msg_dump_walk(FULL, msg);
		CHECK_FCT_DO( msg_free( msg, 1 ), /* continue */ );
	}
	
	
	/* Destroy the message queues */
	CHECK_FCT_DO(  meq_del(&p->p_in_q), /* continue */  );
	CHECK_FCT_DO(  meq_del(&p->p_out_q), /* continue */  );
	
	/* Destroy the initial information of the peer */
	if (p->p_addinfo.pa_flags & PA_FREE)
		free(p->p_addinfo.pa_cbparm);
	
	/* Destroy the socket if not done yet */
	if (p->p_sock) {
		CHECK_SYS_DO( shutdown(p->p_sock, SHUT_RDWR), /* continue */ );
	}
	
	/* Destroy the array of endpoints from CER/CEA */
	free(p->p_peeraddr);
	
	/* Destroy security information */
	while ( ! IS_LIST_EMPTY( &p->p_sec_list ) ) {
		uti_list_t * next = p->p_sec_list.next;
		uti_list_unlink(next);
		free(next);
	}
	
	if (p->p_sec_hdl) {
		CHECK_FCT_DO( sec_modunlink(p->p_sec_hdl), /* continue */ );
	}
	
	free( p->p_remotesec );
	
	/* List of applciations */
	free( p->p_app_list );
	
	/* Product name */
	free( p->p_prodname );
	
	/* Election data if any */
	if (p->p_electdata) {
		(void)_peer_struct_destroy(&p->p_electdata->peer);
		free(p->p_electdata);
	}
	
	/* Finally, free the peer itself */
	free(p);
	
	return 0;
}

/* Finding a peer from a diameter id. diamid must be lowercase! 
 * Upon return, the hash line is locked and must be unlocked.
 * If item is found, it is returned in peer. otherwise, the next element is returned.
 */
int _peer_struct_search(char * diamid, size_t diamidlen, uint32_t * hash, uti_list_t ** peerglob)
{
	uint32_t h;
	size_t   l;
	uti_list_t *li;
	int found = 0;
	
	TRACE_ENTRY("%p %d %p %p", diamid, diamidlen, hash, peerglob);
	
	if (diamidlen == 0)
		l = strlen(diamid);
	else
		l = diamidlen;
	
	h = uti_hash(diamid, l);
	if (hash)
		*hash = h;
	
	/* Now search the peer in the appropriate hash line */
	CHECK_POSIX(  pthread_mutex_lock( H_LOCK(h) )  );
	
	/* Search the line for this peer. A line is ordered by hash and diamid */
	for (li = H_ALL(h)->next; li != H_ALL(h); li = li->next) {
		int cmp = 0;
		_peer_t * cur = _P(li->o);
		
		if (cur->p_hash < h)
			continue;
		
		if (cur->p_hash > h)
			break;
		
		cmp = strncmp(cur->p_diamid, diamid, l);
		
		if (cmp < 0)
			continue;
		
		if (cmp == 0)
			found = 1;
		
		break;
	}
	
	/* return the results */
	if (peerglob)
		*peerglob = li;
	
	return found ? 0 : ENOENT;
}

/* Find a structure from diamid and hash */
peer_t * peer_struct_getptr(char * diamid, uint32_t hash)
{
	uti_list_t *li;
	peer_t * res = NULL;
	
	TRACE_ENTRY("%p %u", diamid, hash);
	
	/* Now search the peer in the appropriate hash line */
	CHECK_POSIX_DO(  pthread_mutex_lock( H_LOCK(hash) ), goto error  );
	
	/* Search the line for this peer. A line is ordered by hash and diamid */
	for (li = H_ALL(hash)->next; li != H_ALL(hash); li = li->next) {
		int cmp = 0;
		_peer_t * cur = _P(li->o);
		
		if (cur->p_hash < hash)
			continue;
		
		if (cur->p_hash > hash)
			break;
		
		cmp = strcmp(cur->p_diamid, diamid);
		
		if (cmp < 0)
			continue;
		
		if (cmp == 0)
			res = (peer_t *)cur;
		
		break;
	}
	
	CHECK_POSIX_DO(  pthread_mutex_unlock( H_LOCK(hash) ), goto error  );
	
	return res;
	
error:
	TRACE_DEBUG(INFO, "Peer was not found or an error occurred ...");
	return NULL;
}

/* Create the list of OPEN peers supporting this app or relay, without the rejected peers */		
int peer_struct_list_open( application_id_t app, rt_dpl_t ** list, uti_list_t * rejects )
{
	uti_list_t * src = NULL;
	
	TRACE_ENTRY("%u %p %p", app, list, rejects);
	
	CHECK_PARAMS( list && rejects );
	
	*list = NULL;
	
	/* Lock the list of active peers. This ensures that the peers will not be modified while accessing their data */
	CHECK_POSIX(  pthread_mutex_lock( &_peers_lists_mtx )  );
	
	/* We go through all the active peers */
	for (src = _peers_actives.next; src != &_peers_actives; src = src->next) {
		_peer_t * src_peer = _P(src->o);
		rt_dpl_t * dst = NULL;
		int match = 0;
		
		/* First check if this peer may be candidate */
		if ((app != 0) && (src_peer->p_app_relay == 0)) {
			int i;
			/* check if the peer supports the required application */
			if (src_peer->p_app_size == 0)
				continue;
			ASSERT(src_peer->p_app_list != NULL);
			
			/* The array is ordered (see _peer_struct_appl_add) */
			for (i = 0; i < src_peer->p_app_size; i++)
				if (src_peer->p_app_list[i].a >= app)
					break;
			/* Skip this peer if the corresponding application was NOT found */
			if (src_peer->p_app_list[i].a != app)
				continue;
		}
		
		/* Now browse the (ordered) rejects list to check if the current peer is not in it */
		while (!IS_LIST_EMPTY(rejects)) {
			/* For the format of the rejects list, see create_list_rejects in routing.c */
			rt_reject_t * next = (rt_reject_t *)(rejects->next);
			
			if (next->hash > src_peer->p_hash)
				break;
			
			if (next->hash == src_peer->p_hash) {
				int cmp = 0;
				avp_value_t *next_av = (avp_value_t *)(next->chain.o);
				
				cmp = strncasecmp((char *)next_av->os.data, src_peer->p_diamid, next_av->os.len);
				
				if (cmp > 0)
					break;
				
				if (cmp == 0) {
					size_t l = strlen(src_peer->p_diamid);
					
					if (next_av->os.len > l)
						break;
					
					if (next_av->os.len == l)
						match = 1;
				}
			}
			
			/* At this point, the next element in rejects is either equal (match = 1) or inferior (match = 0) to src_peer, we can delete it. */
			uti_list_unlink(&next->chain);
			free(next);
		}
		
		
		/* Skip the peer if it was found in the rejects list */
		if (match)
			continue;
	
		/* Ok, we have to copy this peer in the destination list */
		CHECK_MALLOC_DO( dst = (rt_dpl_t *) malloc(sizeof(rt_dpl_t)),
			{ 
				CHECK_POSIX_DO(  pthread_mutex_unlock( &_peers_lists_mtx ), /* continue */  );
				return ENOMEM;
			}  );
		
		memset(dst, 0, sizeof(rt_dpl_t));
		
		dst->peer = (peer_t *)src_peer;
		dst->next = *list;
		*list = dst;
	}
	
	/* Unlock the list of active peers, we're done */
	CHECK_POSIX(  pthread_mutex_unlock( &_peers_lists_mtx )  );
	
	/* We can free any remaining entry in reject list */
	while (!IS_LIST_EMPTY(rejects)) {
		uti_list_t * next = rejects->next;
		uti_list_unlink(next);
		free(next);
	}
	
	return 0;
}


/*************************************************************/
/* A dump function for the peer */
void peer_dump(peer_t * peer, int my_debug_level, int what /* 0: 1-line dump; 1: compact; 2: detailed; 3: develop everything */)
{
	_peer_t * p = _P(peer);
	
	if (my_debug_level > g_conf->log_verbose)
		return;
	
	if (!VALIDATE_PEER( p )) {
		log_debug(" @%p: INVALID! (%s)\n", p,
				(p == NULL) ? "null" : 
					((p->p_eyec != PEER_EYEC) ? "eyec" :
						"deleted"));
		return;
	}
	
	switch (what) {
		case 0:
			log_debug(" @%p: %x %s (%s)\n", p, p->p_hash, p->p_diamid ?  p->p_diamid : "(nil)", STATESTR( p->p_state ));
			break;
		
		case 1:
		default: /* to be completed */
			log_debug(" @%p: id:%s st:%s h:%x rlm:%s hbh:%x\n", 
					p, 
					p->p_diamid ?  p->p_diamid : "(nil)",
					STATESTR( p->p_state ), 
					p->p_hash,
					p->p_realm ?  p->p_realm : "(nil)",
					p->p_hbh);
			log_debug("   fl:%08x gl:%p ac:%s psm:%p ts:%u ev:%s\n",
					p->p_flags,
					IS_LIST_EMPTY(&p->p_global) ? NULL : p->p_global.head,
					IS_LIST_EMPTY(&p->p_active) ? "n" : ((p->p_active.head == &_peers_actives) ? "y" : "E"),
					p->p_psm,
					p->p_ts,
					IS_LIST_EMPTY(&p->p_events) ? "n" : "y");
			log_debug("   inq:%p(%d) th:%p outq:%p(%d) th:%p snt:%s\n",
					p->p_in_q, meq_length_noerr(p->p_in_q),
					p->p_in_th,
					p->p_out_q, meq_length_noerr(p->p_out_q),
					p->p_out_th,
					IS_LIST_EMPTY(&p->p_sent) ? "n" : "y");
			/* To be completed ... */
	}
}

/* return the number of peers in global hash and optionaly dump them */
int peer_dump_hash(int my_debug_level)
{
	int cnt = 0, i;
	
	if (my_debug_level <= g_conf->log_verbose)
		log_debug("=============== HASH dump ==============================\n");
	
	for (i = 0; i < sizeof(_peer_hash) / sizeof(_peer_hash[0]); i++) {
		uti_list_t * next;
		
		if (my_debug_level <= g_conf->log_verbose)
			log_debug("==== line %d (%d)\n", i, _peer_hash[i].count);
		
		CHECK_POSIX(  pthread_mutex_lock(&_peer_hash[i].lock)  );
		
		for (next = _peer_hash[i].all.next; next != &_peer_hash[i].all; next = next->next) {
			cnt ++;
			peer_dump(next->o, my_debug_level,g_conf->log_verbose <= FULL ? 0 : 1);
		}
		
		CHECK_POSIX(  pthread_mutex_unlock(&_peer_hash[i].lock)  );
	}
	if (my_debug_level <= g_conf->log_verbose)
		log_debug("=============== /HASH dump ==============================\n");
	
	return cnt;
}
